# shamelessly stolen from ccrisan https://github.com/qtoggle/qtoggleserver-mppsolar/blob/master/qtoggleserver/mppsolar/io.py
import logging
import random

from .baseio import BaseIO
from ..helpers import get_kwargs

log = logging.getLogger("MPP-Solar")


class TestIO(BaseIO):
    def __init__(self, *args, **kwargs) -> None:
        self._test_data = b"(230.0 50.0 0030 42.0 54.0 56.4 46.0 60 0 0 2 0 0 0 0 0 1 1 0 0 1 0 54.0 0 1 000\x9E\x60\r"
        self._counter = 0

    def send_and_receive(self, *args, **kwargs) -> dict:
        full_command = get_kwargs(kwargs, "full_command")
        command_defn = get_kwargs(kwargs, "command_defn")
        if command_defn is not None:
            self._test_data = command_defn["test_responses"][
                random.randrange(len(command_defn["test_responses"]))
            ]
        response = self._test_data
        # response = b"(PI30\x9a\x0b\r"
        log.debug(f"Raw response {response}")
        return response



import logging

from .device import AbstractDevice


log = logging.getLogger("MPP-Solar")


class mppsolar(AbstractDevice):
    def __init__(self, *args, **kwargs) -> None:
        self._classname = "mppsolar"
        super().__init__(*args, **kwargs)



import logging
import math

from .protocol import AbstractProtocol
from .protocol_helpers import decode4ByteHex, decode2ByteHex, crc8


log = logging.getLogger("MPP-Solar")

SOR = bytes.fromhex("55aaeb90")

COMMANDS = {
    "getInfo": {
        "name": "getInfo",
        "command_code": "97",
        "record_type": "3",
        "description": "BLE Device Information inquiry",
        "help": " -- queries the ble device information",
        "type": "QUERY",
        "response": [
            ["hex", 4, "Header", ""],
            ["hex", 1, "Record Type", ""],
            ["int", 1, "Record Counter", ""],
            ["ascii", 10, "Device Model", ""],
            ["ascii", 10, "Hardware Version", ""],
            ["ascii", 10, "Software Version", ""],
            ["discard", 10, "", ""],
            ["ascii", 16, "Device Name", ""],
            ["ascii", 10, "Device Passcode", ""],
            ["ascii", 14, "Unknown1", ""],
            ["ascii", 14, "Unknown2", ""],
            ["ascii", 16, "User Data", ""],
            ["ascii", 16, "Settings Passcode?", ""],
        ],
        "test_responses": [
            bytes.fromhex(
                "55aaeb9003f14a4b2d42324132345300000000000000332e300000000000332e322e330000000876450004000000506f7765722057616c6c203100000000313233340000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000c2"
            ),
            bytes.fromhex(
                "55aaeb9003b54a4b2d42443641323053313050000000342e300000000000342e312e37000000541d1600040000004e6f7468696e67204a4b31000000000031323334000000000000000000000000323030373038000032303036323834303735000000000000496e707574205573657264617461000031323334353600000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000c4"
            ),
        ],
        "regex": "",
    },
}


class jkAbstractProtocol(AbstractProtocol):
    """
    JKAbstractProtocol - Abstract Handler for JKBMS communication
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"JK02"
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = [
            "",
        ]
        self.SETTINGS_COMMANDS = [
            "getInfo",
        ]
        self.DEFAULT_COMMAND = "getInfo"

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different for JK
        """
        # getInfo = b'\xaa\x55\x90\xeb\x97\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11'
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        log.debug(f"self._command = {self._command}, self._command_defn = {self._command_defn}")
        # End of required variables setting
        if self._command_defn is None:
            # Maybe return a default here?
            log.debug("No command_defn found")
            return None
        if "command_code" in self._command_defn:
            # full command is 20 bytes long
            cmd = bytearray(20)
            # starts with \xaa\x55\x90\xeb
            cmd[0:4] = bytes.fromhex("aa5590eb")
            log.debug(f"cmd with SOR: {cmd}")
            # then has command code
            cmd[4] = int(self._command_defn["command_code"], 16)
            log.debug(f"cmd with command code: {cmd}")
            cmd[-1] = crc8(cmd)
            log.debug(f"cmd with crc: {cmd}")
            return cmd
        return None

    def get_command_defn(self, command):
        log.debug(f"JkBLE get_command_defn for: {command}")
        if command is None:
            log.debug("get_command_defn: command is None")
            return None
        return super().get_command_defn(command)

    def get_responses(self, response):
        """
        Override the default get_responses as its different for JK
        """
        return bytearray(response)

    def is_record_start(self, record):
        if record.startswith(SOR):
            log.debug("SOR found in record")
            return True
        return False

    def is_record_correct_type(self, record, type):
        if len(record) < len(SOR):
            return False
        if record[len(SOR)] == int(type):
            log.debug(f"Record is type {type}")
            return True
        return False

    def is_record_complete(self, record):
        """"""
        # check record starts with 'SOR'
        if not self.is_record_start(record):
            log.debug("No SOR found in record looking for completeness")
            return False
        # check that length one of the valid lengths (300, 320)
        if len(record) == 300 or len(record) == 320:
            # check the crc/checksum is correct for the record data
            crc = ord(record[-1:])
            calcCrc = crc8(record[:-1])
            # print (crc, calcCrc)
            if crc == calcCrc:
                log.debug("Record CRC is valid")
                return True
        return False

    def decode(self, response, command) -> dict:
        msgs = {}
        log.info(f"response passed to decode: {response}")
        # No response
        if response is None:
            log.info("No response")
            msgs["ERROR"] = ["No response", ""]
            return msgs

        # Raw response
        _response = b""
        for item in response:
            _response += chr(item).encode()
        raw_response = _response.decode("utf-8")
        msgs["raw_response"] = [raw_response, ""]

        command_defn = self.get_command_defn(command)
        # Add metadata
        msgs["_command"] = command
        if command_defn is not None:
            msgs["_command_description"] = command_defn["description"]

        # Check for a stored command definition
        if not command_defn:
            # No definiution, so just return the data
            len_command_defn = 0
            log.debug(f"No definition for command {command}, raw response returned")
            msgs["ERROR"] = [
                f"No definition for command {command} in protocol {self._protocol_id}",
                "",
            ]
        else:
            len_command_defn = len(command_defn["response"])
            # Decode response based on stored command definition
            responses = self.get_responses(response)
            log.debug(f"Length of responses {len(responses)}")

            for defn in command_defn["response"]:
                log.debug(f"Processing defn {defn}")
                # ["hex", 4, "Header", ""]
                # if defn[2] == "":
                #    log.debug(f"skipping {defn} and no name defined")
                if defn[0] == "lookup":
                    # calculated lookup column
                    # looks for a columnb based on another
                    # ["int+", 1, "Highest Cell", ""],
                    # ["lookup", "Highest Cell", "Voltage Cell", "Highest Cell Voltage" ],
                    log.debug("using lookup defn")
                    lookup_value = msgs[defn[1]][0]
                    key_tofind = f"{defn[2]}{lookup_value:02d}"
                    value = msgs[key_tofind][0]
                    unit = msgs[key_tofind][1]
                    msgs[defn[3]] = [value, unit]
                elif defn[0] == "hex":
                    log.debug("hex defn")
                    value = ""
                    for x in range(defn[1]):
                        value += f"{responses.pop(0):02x}"
                    if defn[2] != "":
                        msgs[defn[2]] = [value, defn[3]]
                elif defn[0] == "ascii":
                    log.debug("ascii defn")
                    value = ""
                    for x in range(defn[1]):
                        b = responses.pop(0)
                        if b == 0:
                            continue
                        value += f"{b:c}"
                    msgs[defn[2]] = [value, defn[3]]
                elif defn[0] == "uptime":
                    log.debug("uptime defn")
                    value = 0
                    for x in range(defn[1]):
                        b = responses.pop(0)
                        value += b * 256 ** x
                        log.debug(f"Uptime int value {value} for pos {x}")
                    daysFloat = value / (60 * 60 * 24)
                    days = math.trunc(daysFloat)
                    hoursFloat = (daysFloat - days) * 24
                    hours = math.trunc(hoursFloat)
                    minutesFloat = (hoursFloat - hours) * 60
                    minutes = math.trunc(minutesFloat)
                    secondsFloat = (minutesFloat - minutes) * 60
                    seconds = round(secondsFloat)
                    uptime = f"{days}D{hours}H{minutes}M{seconds}S"
                    log.info(f"Uptime result {uptime}")
                    msgs[defn[2]] = [uptime, defn[3]]
                elif defn[0] == "discard":
                    log.debug(f"Discarding {defn[1]} values")
                    value = ""
                    for x in range(defn[1]):
                        value += f"{responses.pop(0):02x}"
                    log.debug(f"Discarded {value}")
                    if defn[2] != "":
                        msgs[defn[2]] = [value, defn[3]]
                elif defn[0] == "int":
                    log.debug("int defn")
                    msgs[defn[2]] = [responses.pop(0), defn[3]]
                elif defn[0] == "int+":
                    log.debug("int+ defn")
                    msgs[defn[2]] = [responses.pop(0) + 1, defn[3]]
                elif defn[0] == "16Int":
                    log.debug("16Int defn")
                    value = responses.pop(0) * 256
                    value += responses.pop(0)
                    # print(f"value {value}")
                    msgs[defn[2]] = [value, defn[3]]
                elif defn[0] == "16Int100":
                    log.debug("16Int100 defn")
                    value = responses.pop(0) * 256
                    value += responses.pop(0)
                    value = value / 100
                    # print(f"value {value}")
                    msgs[defn[2]] = [value, defn[3]]
                elif defn[0] == "16Int1000":
                    log.debug("16Int1000 defn")
                    value = responses.pop(0) * 256
                    value += responses.pop(0)
                    value = value / 1000
                    # print(f"value {value}")
                    msgs[defn[2]] = [value, defn[3]]
                elif defn[0] == "2ByteHex":
                    log.debug("2ByteHex defn")
                    v = responses[:2]
                    responses = responses[2:]
                    value = decode2ByteHex(v)
                    msgs[defn[2]] = [value, defn[3]]
                elif defn[0] == "2ByteHexU":
                    # used for unknow values provides undecoded hex as well
                    log.debug("2ByteHexU defn")
                    v = responses[:2]
                    responses = responses[2:]
                    value = decode2ByteHex(v)
                    msgs[defn[2]] = [value, f"{v[0]:02x} {v[1]:02x}"]
                elif defn[0] == "2ByteHexC":
                    # temperatures seem to be deocded value * 100
                    log.debug("2ByteHexC defn")
                    v = responses[:2]
                    responses = responses[2:]
                    value = decode2ByteHex(v) * 100
                    msgs[defn[2]] = [value, defn[3]]
                elif defn[0] == "4ByteHex":
                    log.debug("4ByteHex defn")
                    v = responses[:4]
                    responses = responses[4:]
                    value = decode4ByteHex(v)
                    msgs[defn[2]] = [value, defn[3]]
                elif defn[0] == "4ByteHexU":
                    log.debug("4ByteHexU defn")
                    v = responses[:4]
                    responses = responses[4:]
                    value = decode4ByteHex(v)
                    msgs[defn[2]] = [value, f"{v[0]:02x} {v[1]:02x} {v[2]:02x} {v[3]:02x}"]
                elif defn[0] == "loop":
                    log.debug("loop defn")
                    # loop of repeating data, eg cell voltages
                    for x in range(defn[1]):
                        param = f"{defn[2]}{x+1:02d}"
                        if defn[4] == "4ByteHex":
                            v = responses[:4]
                            responses = responses[4:]
                            value = decode4ByteHex(v)
                            msgs[param] = [value, defn[3]]
                        if defn[4] == "2ByteHex":
                            v = responses[:2]
                            responses = responses[2:]
                            value = decode2ByteHex(v)
                            msgs[param] = [value, defn[3]]
                        if defn[4] == "16Int1000":
                            value = responses.pop(0) * 256
                            value += responses.pop(0)
                            value = value / 1000
                            msgs[param] = [value, defn[3]]
                elif defn[0] == "rem":
                    log.debug("remainder")
                    msgs["remainder"] = [str(responses), ""]
                    msgs["len remainder"] = [len(responses), ""]
                    return msgs
                else:
                    log.error("undefined type")
        return msgs



#!/usr/bin/env python3
import logging

from binascii import unhexlify
from struct import unpack

log = logging.getLogger("MPP-Solar")


def crc8(byteData):
    """
    Generate 8 bit CRC of supplied string
    """
    CRC = 0
    # for j in range(0, len(str),2):
    for b in byteData:
        # char = int(str[j:j+2], 16)
        # print(b)
        CRC = CRC + b
        CRC &= 0xFF
    return CRC


def decode2ByteHex(hexToDecode):
    """
    Code a 2 byte hexString to volts as per jkbms approach (blackbox determined)
    - need to decode as 4 hex chars
    """
    # hexString = bytes.fromhex(hexToDecode)
    hexString = hexToDecode
    log.debug(f"hexString: {hexString}")

    answer = 0.0

    # Make sure supplied String is long enough
    if len(hexString) != 2:
        log.warning(f"Hex encoded value must be 2 bytes long. Was {len(hexString)} length")
        return 0

    # 1st position
    pos1 = hexString[0] >> 4
    answer += pos1 * (2 ** 4 / 1000)
    log.debug(f"answer after pos1 {answer}")
    # 2nd position
    pos2 = hexString[0] & 0x0F
    answer += pos2 * (2 ** 0 / 1000)
    log.debug(f"answer after pos2 {answer}")
    # 3rd position
    pos3 = hexString[1] >> 4
    answer += pos3 * (2 ** 12 / 1000)
    log.debug(f"answer after pos3 {answer}")
    # 4th position
    pos4 = hexString[1] & 0x0F
    answer += pos4 * (2 ** 8 / 1000)
    log.debug(f"answer after pos4 {answer}")
    log.info(f"Hex {hexString} 2 byte decoded to {answer}")

    return answer


def decode4ByteHex(hexToDecode):
    """
    Code a 4 byte hexString to volts as per jkbms approach (blackbox determined)
    """
    # hexString = decode2ByteHex
    hexString = hexToDecode
    log.debug(f"hexString: {hexString}")

    answer = 0.0

    # Make sure supplied String is long enough
    if len(hexString) != 4:
        log.warning(f"Hex encoded value must be 4 bytes long. Was {len(hexString)} length")
        return 0

    # Use python tools for decode
    answer = unpack("<f", hexString)[0]
    log.info(f"Hex {hexString} 4 byte decoded to {answer}")

    return answer


def crcPI(data_bytes):
    """
    Calculates CRC for supplied data_bytes
    """
    # assert type(byte_cmd) == bytes
    log.debug(f"Calculating CRC for {data_bytes}")

    crc = 0
    da = 0
    crc_ta = [
        0x0000,
        0x1021,
        0x2042,
        0x3063,
        0x4084,
        0x50A5,
        0x60C6,
        0x70E7,
        0x8108,
        0x9129,
        0xA14A,
        0xB16B,
        0xC18C,
        0xD1AD,
        0xE1CE,
        0xF1EF,
    ]

    for c in data_bytes:
        # log.debug('Encoding %s', c)
        # todo fix response for older python
        if type(c) == str:
            c = ord(c)
        da = ((crc >> 8) & 0xFF) >> 4
        crc = (crc << 4) & 0xFFFF

        index = da ^ (c >> 4)
        crc ^= crc_ta[index]

        da = ((crc >> 8) & 0xFF) >> 4
        crc = (crc << 4) & 0xFFFF

        index = da ^ (c & 0x0F)
        crc ^= crc_ta[index]

    crc_low = crc & 0xFF
    crc_high = (crc >> 8) & 0xFF

    if crc_low == 0x28 or crc_low == 0x0D or crc_low == 0x0A:
        crc_low += 1
    if crc_high == 0x28 or crc_high == 0x0D or crc_high == 0x0A:
        crc_high += 1

    crc = crc_high << 8
    crc += crc_low

    log.debug(f"Generated CRC {crc_high:#04x} {crc_low:#04x} {crc:#06x}")
    return [crc_high, crc_low]



import logging
import re
import paho.mqtt.publish as publish

from .baseoutput import baseoutput
from ..helpers import get_kwargs
from ..helpers import key_wanted

log = logging.getLogger("MPP-Solar")


class mqtt(baseoutput):
    def __str__(self):
        return "outputs the to the supplied mqtt broker: eg 'tag'/status/total_output_active_power/value 1250"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"processor.mqtt __init__ kwargs {kwargs}")

    def build_msgs(self, *args, **kwargs):
        data = get_kwargs(kwargs, "data")
        tag = get_kwargs(kwargs, "tag")
        keep_case = get_kwargs(kwargs, "keep_case")
        # Build array of mqtt messages
        msgs = []
        # Remove command and _command_description
        cmd = data.pop("_command", None)
        data.pop("_command_description", None)
        data.pop("raw_response", None)
        if tag is None:
            tag = cmd
        # Loop through responses
        for key in data:
            value = data[key][0]
            unit = data[key][1]
            # remove spaces
            key = key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, filter, excl_filter):
                log.debug(f"tag {tag}, key {key}, value {value}, unit {unit}")
                # 'tag'/status/total_output_active_power/value 1250
                # 'tag'/status/total_output_active_power/unit W
                msg = {"topic": f"{tag}/status/{key}/value", "payload": value}
                msgs.append(msg)
                if unit:
                    msg = {"topic": f"{tag}/status/{key}/unit", "payload": unit}
                    msgs.append(msg)
        log.debug(msgs)
        return msgs

    def output(self, *args, **kwargs):
        log.info("Using output processor: mqtt")
        log.debug(f"processor.mqtt.output kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        if data is None:
            return
        mqtt_broker = get_kwargs(kwargs, "mqtt_broker", "localhost")
        mqtt_port = get_kwargs(kwargs, "mqtt_port", 1883)
        mqtt_user = get_kwargs(kwargs, "mqtt_user")
        mqtt_pass = get_kwargs(kwargs, "mqtt_pass")

        filter = get_kwargs(kwargs, "filter")
        if filter is not None:
            filter = re.compile(filter)
        excl_filter = get_kwargs(kwargs, "excl_filter")
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        if mqtt_user is not None and mqtt_pass is not None:
            auth = {"username": mqtt_user, "password": mqtt_pass}
            log.info(f"Using mqtt authentication, username: {mqtt_user}, password: [supplied]")
        else:
            log.debug("No mqtt authentication used")
            auth = None

        msgs = self.build_msgs(**kwargs)
        log.debug(f"mqtt.output msgs {msgs}")
        if msgs:
            publish.multiple(msgs, hostname=mqtt_broker, port=mqtt_port, auth=auth)
        else:
            log.warn("MQTT build_msgs returned no messages")



import logging
import re

from .baseoutput import baseoutput
from ..helpers import get_kwargs
from ..helpers import key_wanted

log = logging.getLogger("MPP-Solar")


class screen(baseoutput):
    def __str__(self):
        return "[the default output module] outputs the results to standard out in a slightly formatted way"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"processor.screen __init__ kwargs {kwargs}")

    def output(self, *args, **kwargs):
        log.info("Using output processor: screen")
        log.debug(f"processor.screen.output kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        if data is None:
            return

        keep_case = get_kwargs(kwargs, "keep_case")
        filter = get_kwargs(kwargs, "filter")
        if filter is not None:
            filter = re.compile(filter)
        excl_filter = get_kwargs(kwargs, "excl_filter")
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        _desc = "No description found"
        if "_command_description" in data:
            _desc = data["_command_description"]
            del data["_command_description"]
        if "_command" in data:
            print(f"Command: {data['_command']} - {_desc}")
            print("-" * 60)
            del data["_command"]
        if "raw_response" in data:
            del data["raw_response"]

        print(f"{'Parameter':<30}\t{'Value':<15} Unit")
        for key in data:
            value = data[key][0]
            unit = data[key][1]
            # remove spaces
            key = key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, filter, excl_filter):
                print(f"{key:<30}\t{value:<15}\t{unit:<4}")



import json as js
import logging
import re

from .baseoutput import baseoutput
from ..helpers import get_kwargs
from ..helpers import key_wanted

log = logging.getLogger("MPP-Solar")


class json(baseoutput):
    def __str__(self):
        return "outputs the results to standard out in json format"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"processor.json __init__ kwargs {kwargs}")

    def output(self, *args, **kwargs):
        log.info("Using output processor: json")
        log.debug(f"processor.json.output kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        keep_case = get_kwargs(kwargs, "keep_case")
        data.pop("raw_response", None)

        filter = get_kwargs(kwargs, "filter")
        if filter is not None:
            filter = re.compile(filter)
        excl_filter = get_kwargs(kwargs, "excl_filter")
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        output = {}
        for key in data:
            value = data[key]
            if isinstance(value, list):
                value = data[key][0]
            # unit = data[key][1]
            # remove spaces
            key = key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, filter, excl_filter):
                output[key] = value

        print(js.dumps(output))



import abc
import logging
import re
from typing import Tuple

from .protocol_helpers import crcPI as crc

log = logging.getLogger("MPP-Solar")


class AbstractProtocol(metaclass=abc.ABCMeta):
    def __init__(self, *args, **kwargs) -> None:
        self._command = None
        self._command_dict = None
        self.COMMANDS = {}
        self.STATUS_COMMANDS = None
        self.SETTINGS_COMMANDS = None
        self.DEFAULT_COMMAND = None
        self._protocol_id = None

    def get_protocol_id(self) -> bytes:
        return self._protocol_id

    def get_full_command(self, command) -> bytes:
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")

        byte_cmd = bytes(command, "utf-8")
        # calculate the CRC
        crc_high, crc_low = crc(byte_cmd)
        # combine byte_cmd, CRC , return
        full_command = byte_cmd + bytes([crc_high, crc_low, 13])
        log.debug(f"full command: {full_command}")
        return full_command

    def get_command_defn(self, command) -> dict:
        log.debug(f"get_command_defn for: {command}")
        if command in self.COMMANDS:
            # print(command)
            log.debug(f"Found command {command} in protocol {self._protocol_id}")
            return self.COMMANDS[command]
        for _command in self.COMMANDS:
            if "regex" in self.COMMANDS[_command] and self.COMMANDS[_command]["regex"]:
                log.debug(f"Regex commands _command: {_command}")
                _re = re.compile(self.COMMANDS[_command]["regex"])
                match = _re.match(command)
                if match:
                    log.debug(
                        f"Matched: {command} to: {self.COMMANDS[_command]['name']} value: {match.group(1)}"
                    )
                    self._command_value = match.group(1)
                    return self.COMMANDS[_command]
        log.info(f"No command_defn found for {command}")
        return None

    def get_responses(self, response) -> list:
        """
        Default implementation of split and trim
        """
        # Trim leading '(' + trailing CRC and \r of response, then split
        if type(response) is str:
            return response[1:-3].split(" ")
        return response[1:-3].split(b" ")

    def check_response_valid(self, response) -> Tuple[bool, dict]:
        """
        Simplest validity check, CRC checks should be added to individual protocols
        """
        if response is None:
            return False, {"ERROR": ["No response", ""]}
        return True, {}

    def decode(self, response, command) -> dict:
        log.info(f"response passed to decode: {response}")

        valid, msgs = self.check_response_valid(response)
        if not valid:
            log.info(msgs["ERROR"][0])
            return msgs

        # Add Raw response
        _response = b""
        for item in response:
            if type(item) is int:
                _response += chr(item).encode()
            else:
                _response += item.encode()
        raw_response = _response.decode("utf-8")
        msgs["raw_response"] = [raw_response, ""]

        command_defn = self.get_command_defn(command)
        # Add metadata
        msgs["_command"] = command
        if command_defn is not None:
            msgs["_command_description"] = command_defn["description"]

        # Check for a stored command definition
        if not command_defn:
            # No definiution, so just return the data
            len_command_defn = 0
            log.debug(f"No definition for command {command}, (splitted) raw response returned")
            msgs["WARNING"] = [
                f"No definition for command {command} in protocol {self._protocol_id}",
                "",
            ]
        else:
            len_command_defn = len(command_defn["response"])
        # Decode response based on stored command definition
        # if not self.is_response_valid(response):
        #    log.info('Invalid response')
        #    msgs['ERROR'] = ['Invalid response', '']
        #    msgs['response'] = [response, '']
        #    return msgs

        responses = self.get_responses(response)

        log.debug(f"trimmed and split responses: {responses}")

        for i, result in enumerate(responses):
            # decode result
            if type(result) is bytes:
                result = result.decode("utf-8")
            # Check if we are past the 'known' responses
            if i >= len_command_defn:
                resp_format = ["string", f"Unknown value in response {i}", ""]
            else:
                resp_format = command_defn["response"][i]

            # key = "{}".format(resp_format[1]).lower().replace(" ", "_")
            key = resp_format[1]
            # log.debug(f'result {result}, key {key}, resp_format {resp_format}')
            # Process results
            if resp_format[0] == "float":
                if "--" in result:
                    result = 0
                msgs[key] = [float(result), resp_format[2]]
            elif resp_format[0] == "int":
                if "--" in result:
                    result = 0
                msgs[key] = [int(result), resp_format[2]]
            elif resp_format[0] == "string":
                msgs[key] = [result, resp_format[2]]
            elif resp_format[0] == "10int":
                if "--" in result:
                    result = 0
                msgs[key] = [float(result) / 10, resp_format[2]]
            # eg. ['option', 'Output source priority', ['Utility first', 'Solar first', 'SBU first']],
            elif resp_format[0] == "option":
                msgs[key] = [resp_format[2][int(result)], ""]
            # eg. ['keyed', 'Machine type', {'00': 'Grid tie', '01': 'Off Grid', '10': 'Hybrid'}],
            elif resp_format[0] == "keyed":
                msgs[key] = [resp_format[2][result], ""]
            # eg. ['flags', 'Device status', [ 'is_load_on', 'is_charging_on' ...
            elif resp_format[0] == "flags":
                for j, flag in enumerate(result):
                    msgs[resp_format[2][j]] = [int(flag), "True - 1/False - 0"]
            # eg. ['stat_flags', 'Warning status', ['Reserved', 'Inver...
            elif resp_format[0] == "stat_flags":
                output = ""
                for j, flag in enumerate(result):
                    if flag == "1":
                        output = "{}\n\t- {}".format(output, resp_format[2][j])
                msgs[key] = [output, ""]
            # eg. ['enflags', 'Device Status', {'a': {'name': 'Buzzer', 'state': 'disabled'},
            elif resp_format[0] == "enflags":
                # output = {}
                status = "unknown"
                for item in result:
                    if item == "E":
                        status = "enabled"
                    elif item == "D":
                        status = "disabled"
                    else:
                        # output[resp_format[2][item]['name']] = status
                        # _key = "{}".format(resp_format[2][item]["name"]).lower().replace(" ", "_")
                        if resp_format[2].get(item, None):
                            _key = resp_format[2][item]["name"]
                        else:
                            _key = "unknown_{}".format(item)
                        msgs[_key] = [status, ""]
                # msgs[key] = [output, '']
            elif command_defn["type"] == "SETTER":
                # _key = "{}".format(command_defn["name"]).lower().replace(" ", "_")
                _key = command_defn["name"]
                msgs[_key] = [result, ""]
            else:
                msgs[i] = [result, ""]
        return msgs



import logging

from .protocol import AbstractProtocol
from .protocol_helpers import crcPI as crc

# from .pi30 import COMMANDS

log = logging.getLogger("MPP-Solar")

COMMANDS = {
    "ET": {
        "name": "ET",
        "prefix": "^P005",
        "description": "Total Generated Energy query",
        "help": " -- Query total generated energy",
        "type": "QUERY",
        "response": [["int", "Total generated energy", "KWh"]],
        "test_responses": [
            b"",
        ],
        "regex": "",
    },
    "GS": {
        "name": "GS",
        "prefix": "^P005",
        "description": "General status query",
        "help": " -- Query general status information",
        "type": "QUERY",
        "response": [
            ["10int", "Grid voltage", "V"],
            ["10int", "Grid frequency", "Hz"],
            ["10int", "AC output voltage", "V"],
            ["10int", "AC output frequency", "Hz"],
            ["int", "AC output apparent power", "VA"],
            ["int", "AC output active power", "W"],
            ["int", "Output load percent", "%"],
            ["10int", "Battery voltage", "V"],
            ["10int", "Battery voltage from SCC", "V"],
            ["10int", "Battery voltage from SCC2", "V"],
            ["int", "Battery discharge current", "A"],
            ["int", "Battery charging current", "A"],
            ["int", "Battery capacity", "%"],
            ["int", "Inverter heat sink temperature", "°C"],
            ["int", "MPPT1 charger temperature", "°C"],
            ["int", "MPPT2 charger temperature", "°C"],
            ["int", "PV1 Input power", "W"],
            ["int", "PV2 Input power", "W"],
            ["10int", "PV1 Input voltage", "V"],
            ["10int", "PV2 Input voltage", "V"],
            [
                "option",
                "Setting value configuration state",
                ["Nothing changed", "Something changed"],
            ],
            [
                "option",
                "MPPT1 charger status",
                ["abnormal", "normal but not charged", "charging"],
            ],
            [
                "option",
                "MPPT2 charger status",
                ["abnormal", "normal but not charged", "charging"],
            ],
            ["option", "Load connection", ["disconnect", "connect"]],
            ["option", "Battery power direction", ["donothing", "charge", "discharge"]],
            ["option", "DC/AC power direction", ["donothing", "AC-DC", "DC-AC"]],
            ["option", "Line power direction", ["donothing", "input", "output"]],
            ["int", "Local parallel ID", ""],
        ],
        "test_responses": [
            b"D1062232,499,2232,499,0971,0710,019,008,000,000,000,000,000,044,000,000,0520,0000,1941,0000,0,2,0,1,0,2,1,0\x09\x7b\r",
        ],
        "regex": "",
    },
    "MOD": {
        "name": "MOD",
        "prefix": "^P006",
        "description": "Working mode query",
        "help": " -- Query the working mode",
        "type": "QUERY",
        "response": [
            [
                "option",
                "Working mode",
                [
                    "Power on mode",
                    "Standby mode",
                    "Bypass mode",
                    "Battery mode",
                    "Fault mode",
                    "Hybrid mode(Line mode, Grid mode)",
                ],
            ]
        ],
        "test_responses": [
            b"",
        ],
        "regex": "",
    },
    "PI": {
        "name": "PI",
        "prefix": "^P005",
        "description": "Device Protocol Version inquiry",
        "help": " -- queries the device protocol version",
        "type": "QUERY",
        "response": [["string", "Protocol Version", ""]],
        "test_responses": [
            b"",
        ],
        "regex": "",
    },
}


class pi18(AbstractProtocol):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI18"
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = ["MOD", "GS", "ET"]
        self.SETTINGS_COMMANDS = [
            "PI",
        ]
        self.DEFAULT_COMMAND = "PI"

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different for PI18
        """
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # End of required variables setting
        if self._command_defn is None:
            return None

        _cmd = bytes(self._command, "utf-8")
        _type = self._command_defn["type"]

        # calculate the CRC
        crc_high, crc_low = crc(_cmd)
        # combine byte_cmd, CRC , return
        # PI18 full command "^P005GS\x..\x..\r"
        command_crc = _cmd + bytes([crc_high, crc_low, 13])
        if _type == "QUERY":
            _prefix = f"^P{len(command_crc):03}"
            full_command = bytes(_prefix, "utf-8") + command_crc
        log.debug(f"full command: {full_command}")
        return full_command

    def get_responses(self, response):
        """
        Override the default get_responses as its different for PI18
        """
        responses = response.split(b",")
        # Drop ^Dxxx from first response
        responses[0] = responses[0][4:]
        # Remove CRC of last response
        responses[-1] = responses[-1][:-3]
        return responses



#!/usr/bin/env python3
import logging
import re

log = logging.getLogger("MPP-Solar")


def get_kwargs(kwargs, key, default=None):
    if not key in kwargs or not kwargs[key]:
        return default
    return kwargs[key]


def key_wanted(key, filter=None, excl_filter=None):
    # remove any specifically excluded keys
    if excl_filter is not None and excl_filter.search(key):
        log.debug(f"key {key} matches excl_filter {excl_filter} so key excluded")
        return False
    if filter is None:
        log.info(f"No filter and key {key} not excluded by excl_filter {excl_filter} so key wanted")
        return True
    elif filter.search(key):
        log.info(
            f"key {key} matches filter {filter} and not excl_filter {excl_filter} so key wanted"
        )
        return True
    else:
        log.debug(f"key {key} does not match filter {filter} so key excluded")
        return False



# shamelessly stolen from ccrisan https://github.com/qtoggle/qtoggleserver-mppsolar/blob/master/qtoggleserver/mppsolar/io.py
import abc
import logging

# from time import sleep
log = logging.getLogger("MPP-Solar")


class BaseIO(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def send_and_receive(self, *args, **kwargs) -> dict:
        raise NotImplementedError



import json as js
import logging
import paho.mqtt.publish as publish
import paho.mqtt.subscribe as subscribe
import paho.mqtt.client as mqttc

import time

from .baseio import BaseIO
from ..helpers import get_kwargs

log = logging.getLogger("MPP-Solar")


class MqttIO(BaseIO):
    def __init__(self, *args, **kwargs) -> None:
        # self._serial_port = device_path
        # self._serial_baud = serial_baud
        self.mqtt_broker = get_kwargs(kwargs, "mqtt_broker", "localhost")
        self.mqtt_port = get_kwargs(kwargs, "mqtt_port", 1883)
        self.mqtt_user = get_kwargs(kwargs, "mqtt_user")
        self.mqtt_pass = get_kwargs(kwargs, "mqtt_pass")
        self.client_id = get_kwargs(kwargs, "client_id")
        log.info(
            f"MqttIO.__init__ client_id: {self.client_id},  mqtt_broker: {self.mqtt_broker}, port: {self.mqtt_port}, user: {self.mqtt_user}, pass: {self.mqtt_pass}"
        )
        self._msg = None

    def sub_cb(self, client, userdata, message):
        log.debug(f"Mqttio sub_cb got msg, topic: {message.topic}, payload: {message.payload}")
        self._msg = message

    def send_and_receive(self, *args, **kwargs) -> dict:
        full_command = get_kwargs(kwargs, "full_command")
        command = get_kwargs(kwargs, "command")
        client_id = self.client_id

        wait_time = 5
        response_line = None
        command_topic = f"{client_id}/command"
        result_topic = f"{client_id}/result"
        # print(self.mqtt_broker)
        # Create mqtt client
        # Client(client_id="", clean_session=True, userdata=None, protocol=MQTTv311, transport="tcp")
        mqtt_client = mqttc.Client()
        # mqtt_client.on_connect = on_connect

        if self.mqtt_user is not None and self.mqtt_pass is not None:
            # auth = {"username": self.mqtt_user, "password": self.mqtt_pass}
            log.info(f"Using mqtt authentication, username: {self.mqtt_user}, password: [supplied]")
            mqtt_client.username_pw_set(self.mqtt_user, password=self.mqtt_pass)
        else:
            log.debug("No mqtt authentication used")
            # auth = None

        # connect(host, port=1883, keepalive=60, bind_address="")
        mqtt_client.connect(self.mqtt_broker, port=self.mqtt_port)

        payload = full_command
        _payload = "{'command': '%s', 'full_command': %s}" % (command, full_command)
        print(_payload)

        log.debug(f"Publishing {payload} to topic: {command_topic}")

        # publish(topic, payload=None, qos=0, retain=False)
        mqtt_client.publish(command_topic, payload=payload)

        mqtt_client.on_message = self.sub_cb
        mqtt_client.subscribe(result_topic)
        mqtt_client.loop_start()
        time.sleep(wait_time)
        mqtt_client.loop_stop(force=False)

        if self._msg is None:
            # Didnt get a result
            return {
                "ERROR": [
                    f"Mqtt result message not received on topic {result_topic} after {wait_time}sec",
                    "",
                ]
            }
        else:
            msg_topic = self._msg.topic
            msg_payload = self._msg.payload.decode("unicode_escape")
            log.debug(
                f"mqtt raw response on {self._msg.topic} was: {self._msg.payload}, payload type: {type(self._msg.payload)}"
            )
            # msg_payload = self._msg.payload
            self._msg = None
            log.debug(f"mqtt response on {msg_topic} was: {msg_payload}")
            return msg_payload



# !/usr/bin/python3
from argparse import ArgumentParser
import importlib
import logging
from sys import exit

from .version import __version__, __version_comment__  # noqa: F401

log = logging.getLogger("MPP-Solar")
# logging (DEBUG, INFO, WARNING, ERROR, CRITICAL)
# set default log level
log.setLevel(logging.WARNING)
logging.basicConfig()

# Ongoing effort to map model "numbers" to the correct protocol
MODEL_PROTOCOL_MAP = {
    "standard": "PI30",
    "4048MS": "PI30",
    "LV5048": "PI41",
    "PI18": "PI18",
}


def get_outputs(output_list):
    """
    Take a comma separated list of output names
    attempt to find and instantiate the corresponding module
    return array of modules
    """
    ops = []
    outputs = output_list.split(",")
    for output in outputs:
        log.info(f"attempting to create output processor: {output}")
        try:
            output_module = importlib.import_module("mppsolar.outputs." + output, ".")
            output_class = getattr(output_module, output)
            ops.append(output_class())
        except ModuleNotFoundError:
            # perhaps raise a Powermon exception here??
            # maybe warn and keep going, only error if no outputs found?
            log.critical(f"No module found for output processor {output}")
    return ops


def get_device_class(device_type=None):
    """
    Take a device type string
    attempt to find and instantiate the corresponding module
    return class if found, otherwise return None
    """
    if device_type is None:
        return None
    device_type = device_type.lower()
    try:
        device_module = importlib.import_module("mppsolar.devices." + device_type, ".")
    except ModuleNotFoundError as e:
        # perhaps raise a mppsolar exception here??
        log.critical(f"Error loading device {device_type}: {e}")
        return None
    device_class = getattr(device_module, device_type)
    return device_class


def get_protocol_for_model(model=None):
    """
    Try to find the correct protocol for a given model of inverter
    """
    if model is None:
        return None
    if model in MODEL_PROTOCOL_MAP:
        protocol = MODEL_PROTOCOL_MAP[model]
        log.info(f"Model {model} specified, setting protocol to {protocol}")
        return protocol
    log.info(f"Cannot find protocol for model {model}")
    return None


def main():
    description = f"Solar Device Command Utility, version: {__version__}, {__version_comment__}"
    parser = ArgumentParser(description=description)
    parser.add_argument(
        "-n",
        "--name",
        type=str,
        help="Specifies the device name - used to differentiate different devices",
        default="unnamed",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=str,
        help="Specifies the device communications port (/dev/ttyUSB0 [default], /dev/hidraw0, test, ...)",
        default="/dev/ttyUSB0",
    )
    parser.add_argument(
        "--porttype",
        type=str,
        help="overrides the device communications port type",
        default=None,
    )
    if parser.prog == "jkbms":
        parser.add_argument(
            "-P",
            "--protocol",
            type=str,
            help="Specifies the device command and response protocol, (default: JK04)",
            default="JK04",
            choices=["JK02", "JK04", "JK485"],
        )
    else:
        parser.add_argument(
            "-P",
            "--protocol",
            type=str,
            help="Specifies the device command and response protocol, (default: PI30)",
            default="PI30",
            choices=["PI00", "PI16", "PI18", "PI30", "PI41"],
        )
    parser.add_argument(
        "-T",
        "--tag",
        type=str,
        help="Override the command name and use this instead (for mqtt and influx type output processors)",
    )
    parser.add_argument(
        "-b",
        "--baud",
        type=int,
        help="Baud rate for serial communications (default: 2400)",
        default=2400,
    )
    parser.add_argument(
        "-M",
        "--model",
        type=str,
        help='Specifies the inverter model to select commands for, defaults to "standard", currently supports LV5048',
        default="standard",
    )

    parser.add_argument(
        "-o",
        "--output",
        nargs="?",
        type=str,
        help="Specifies the output processor(s) to use [comma separated if multiple] (screen [default]) leave blank to give list",
        const="help",
        default="screen",
    )
    parser.add_argument(
        "--keepcase",
        action="store_true",
        help="Do not convert the field names to lowercase",
        default=False,
    )
    parser.add_argument(
        "--filter",
        type=str,
        help="Specifies the filter to reduce the output - only those fields that match will be output (uses re.search)",
        default=None,
    )
    parser.add_argument(
        "--exclfilter",
        type=str,
        help="Specifies the filter to reduce the output - any fields that match will be excluded from the output (uses re.search)",
        default=None,
    )
    parser.add_argument(
        "-q",
        "--mqttbroker",
        type=str,
        help="Specifies the mqtt broker to publish to if using a mqtt output (localhost [default], hostname, ip.add.re.ss ...)",
        default="localhost",
    )
    parser.add_argument(
        "--mqttport",
        type=str,
        help="Specifies the mqtt broker port if needed (default: 1883)",
        default=1883,
    )
    parser.add_argument(
        "--mqttuser",
        type=str,
        help="Specifies the username to use for authenticated mqtt broker publishing",
        default=None,
    )
    parser.add_argument(
        "--mqttpass",
        type=str,
        help="Specifies the password to use for authenticated mqtt broker publishing",
        default=None,
    )
    parser.add_argument("-c", "--command", nargs="?", const="help", help="Command to run")
    parser.add_argument(
        "-C",
        "--configfile",
        type=str,
        help="Full location of config file",
        default=None,
    )
    parser.add_argument("--daemon", action="store_true", help="Run as daemon")
    parser.add_argument("--getstatus", action="store_true", help="Get Inverter Status")
    parser.add_argument("--getsettings", action="store_true", help="Get Inverter Settings")

    parser.add_argument("-v", "--version", action="store_true", help="Display the version")
    parser.add_argument(
        "-D",
        "--debug",
        action="store_true",
        help="Enable Debug and above (i.e. all) messages",
    )
    parser.add_argument(
        "-I", "--info", action="store_true", help="Enable Info and above level messages"
    )

    args = parser.parse_args()
    prog_name = parser.prog
    s_prog_name = prog_name.replace("-", "")
    log_name = s_prog_name.upper()

    # Display verison if asked
    log.info(description)
    if args.version:
        print(description)
        exit(0)
    # Turn on debug if needed
    if args.debug:
        log.setLevel(logging.DEBUG)
        # ch.setLevel(logging.DEBUG)
    elif args.info:
        log.setLevel(logging.INFO)
        # ch.setLevel(logging.INFO)

    mqtt_broker = args.mqttbroker
    mqtt_port = args.mqttport
    mqtt_user = args.mqttuser
    mqtt_pass = args.mqttpass
    filter = args.filter
    excl_filter = args.exclfilter
    keep_case = args.keepcase

    _commands = []
    # Initialize Daemon
    if args.daemon:
        import time
        import systemd.daemon

        # Tell systemd that our service is ready
        systemd.daemon.notify("READY=1")
        print("Service Initializing ...")
        # set some default-defaults
        pause = 60

    # If config file specified, process
    if args.configfile:
        import configparser

        log.debug(f"args.configfile is true: {args.configfile}")
        config = configparser.ConfigParser()
        config.read(args.configfile)
        sections = config.sections()
        # Process setup section
        if "SETUP" in config:
            pause = config["SETUP"].getint("pause", fallback=60)
            mqtt_broker = config["SETUP"].get("mqtt_broker", fallback="localhost")
            mqtt_port = config["SETUP"].get("mqtt_port", fallback=1883)
            mqtt_user = config["SETUP"].get("mqtt_user", fallback=None)
            mqtt_pass = config["SETUP"].get("mqtt_pass", fallback=None)
            sections.remove("SETUP")
        # Process 'command' sections
        for section in sections:
            name = section
            protocol = config[section].get("protocol", fallback=None)
            model = config[section].get("model", fallback=None)
            if model is not None and protocol is None:
                protocol = get_protocol_for_model(model)
            type = config[section].get("type", fallback="mppsolar")
            port = config[section].get("port", fallback="/dev/ttyUSB0")
            baud = config[section].get("baud", fallback=2400)
            command = config[section].get("command")
            tag = config[section].get("tag")
            outputs = config[section].get("outputs", fallback="screen")
            porttype = config[section].get("porttype", fallback=None)
            #
            device_class = get_device_class(type)
            log.debug(f"device_class {device_class}")
            # The device class __init__ will instantiate the port communications and protocol classes
            device = device_class(
                name=name,
                port=port,
                protocol=protocol,
                outputs=outputs,
                baud=baud,
                porttype=porttype,
                mqtt_broker=mqtt_broker,
                mqtt_port=mqtt_port,
                mqtt_user=mqtt_user,
                mqtt_pass=mqtt_pass,
            )
            # build array of commands
            _commands.append((device, command, tag, outputs))

            if args.daemon:
                print(f"Config file: {args.configfile}")
                print(f"Config setting - pause: {pause}")
                print(f"Config setting - mqtt_broker: {mqtt_broker}, port: {mqtt_port}")
                print(f"Config setting - command sections found: {len(sections)}")
            else:
                log.info(f"Config file: {args.configfile}")
                log.info(f"Config setting - pause: {pause}")
                log.info(f"Config setting - mqtt_broker: {mqtt_broker}, port: {mqtt_port}")
                log.info(f"Config setting - command sections found: {len(sections)}")

    else:
        # No configfile specified
        # process some arguments
        if args.tag:
            tag = args.tag
        else:
            tag = args.command
        if args.model is not None and args.protocol is None:
            args.protocol = get_protocol_for_model(args.model)

        # create instance of device (supplying port + protocol types)
        log.info(
            f'Creating device "{args.name}" (type: "{s_prog_name}") on port "{args.port} (porttype={args.porttype})" using protocol "{args.protocol}" for command "{args.command}" (tag: {tag})'
        )
        device_class = get_device_class(s_prog_name)
        log.debug(f"device_class {device_class}")
        # The device class __init__ will instantiate the port communications and protocol classes
        device = device_class(
            name=args.name,
            port=args.port,
            protocol=args.protocol,
            baud=args.baud,
            porttype=args.porttype,
            mqtt_broker=mqtt_broker,
            mqtt_port=mqtt_port,
            mqtt_user=mqtt_user,
            mqtt_pass=mqtt_pass,
        )
        #

        # determine whether to run command or call helper function
        commands = []
        if args.command == "help":
            commands.append("list_commands")
        elif args.output == "help":
            commands.append("list_outputs")
            args.output = "screen"
            # print("Available output modules:")
            # for result in results:
            #    print(result)
            # exit()
        elif args.getstatus:
            # use get_status helper
            commands.append("get_status")
            # TODO: implement get_status
        elif args.getsettings:
            # use get_settings helper
            commands.append("get_settings")
            # TODO: implement get_settings
        elif args.command is None:
            # run the command
            commands.append("")
        else:
            for cmd in args.command.split(","):
                commands.append(cmd)

        outputs = args.output
        for command in commands:
            _commands.append((device, command, tag, outputs))
        log.debug(f"Commands {_commands}")

    while True:
        # Loop through the configured commands
        if not args.daemon:
            log.info(f"Looping {len(_commands)} commands")
        for _device, _command, _tag, _outputs in _commands:
            ## for item in mppUtilArray:
            # Tell systemd watchdog we are still alive
            if args.daemon:
                systemd.daemon.notify("WATCHDOG=1")
                print(
                    f"Getting results from device: {_device} for command: {_command}, tag: {_tag}, outputs: {_outputs}"
                )
            else:
                log.info(
                    f"Getting results from device: {_device} for command: {_command}, tag: {_tag}, outputs: {_outputs}"
                )
            results = _device.run_command(command=_command)
            log.debug(f"results: {results}")
            # send to output processor(s)
            outputs = get_outputs(_outputs)
            for op in outputs:
                # maybe include the command and what the command is im the output
                # eg QDI run, Display Inverter Default Settings
                log.debug(f"Using output filter: {filter}")
                op.output(
                    data=results,
                    tag=_tag,
                    mqtt_broker=mqtt_broker,
                    mqtt_port=mqtt_port,
                    mqtt_user=mqtt_user,
                    mqtt_pass=mqtt_pass,
                    topic=prog_name,
                    filter=filter,
                    excl_filter=excl_filter,
                    keep_case=keep_case,
                )
                # Tell systemd watchdog we are still alive
        if args.daemon:
            systemd.daemon.notify("WATCHDOG=1")
            print(f"Sleeping for {pause} sec")
            time.sleep(pause)
        else:
            # Dont loop unless running as daemon
            log.debug("Not daemon, so not looping")
            break



import logging

from .protocol import AbstractProtocol

# from .pi30 import COMMANDS

log = logging.getLogger("MPP-Solar")

# (AAA BBB CCC DDD EEE
# (000 001 002 003 004

COMMANDS = {
    "QPI": {
        "name": "QPI",
        "description": "Device Protocol Version inquiry",
        "help": " -- queries the device protocol version",
        "type": "QUERY",
        "response": [["string", "Protocol Version", ""]],
        "test_responses": [
            b"(PI00\x42\x42\r",
        ],
        "regex": "",
    },
    "QPIRI": {
        "name": "QPIRI",
        "description": "Device rating inquiry",
        "help": " -- queries the Inverter ratings",
        "type": "QUERY",
        "response": [
            ["float", "Grid Input Voltage Rating", "V"],
            ["float", "Grid Input Frequency Rating", "Hz"],
            ["float", "Grid Input Current Rating", "A"],
            ["float", "AC Output Voltage Rating", "V"],
            ["float", "AC Output Current Rating", "A"],
            ["float", "Per MPPT Current Rating", "A"],
            ["float", "Battery Voltage Rating", "V"],
            ["int", "MPPT Track Number", ""],
            [
                "keyed",
                "Machine Type",
                {"00": "Grid tie", "01": "Off Grid", "10": "Hybrid"},
            ],
            ["option", "Topology", ["transformerless", "transformer"]],
        ],
        "test_responses": [
            b"(230.0 50.0 013.0 230.0 013.0 18.0 048.0 1 10 0\x86\x42\r",
        ],
        "regex": "",
    },
}


class pi00(AbstractProtocol):
    """
    PI00 - test protocol handler
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI00"
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = [
            "",
        ]
        self.SETTINGS_COMMANDS = [
            "QPI",
        ]
        self.DEFAULT_COMMAND = "QPI"

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different for PI00
        """
        log.info(
            f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands"
        )
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # End of required variables setting
        cmd = bytes(self._command, "utf-8")
        # combine byte_cmd, return
        full_command = cmd + bytes([13])
        log.debug(f"full command: {full_command}")
        return full_command

    def get_responses(self, response):
        """
        Override the default get_responses as its different for PI00
        """
        responses = response.split(b" ")
        # Trim leading '(' of first response
        responses[0] = responses[0][1:]
        # Remove CRC and \r of last response
        responses[-1] = responses[-1][:-3]
        return responses



import logging

from .pi30 import pi30

log = logging.getLogger('MPP-Solar')

# New / overriden commands
NEW_COMMANDS = {
    'QP2GS': {
        "name": "QP2GS",
        "description": "Parallel Information inquiry",
        "help": " -- example: QP2GS1 queries the values of various metrics from instance 1 of parallel setup Inverters (numbers from 0)",
        "type": "QUERY",
        "response": [
            ["int", "Parallel instance number??", ""],
            ["int", "Serial number", ""],
            ["keyed", "Work mode", {"P": "Power On Mode",
                                    "S": "Standby Mode",
                                    "L": "Line Mode",
                                    "B": "Battery Mode",
                                    "F": "Fault Mode",
                                    "H": "Power Saving Mode"}],
            ["keyed", "Fault code", {"00": "No fault",
                                     "01": "Fan is locked",
                                     "02": "Over temperature",
                                     "03": "Battery voltage is too high",
                                     "04": "Battery voltage is too low",
                                     "05": "Output short circuited or Over temperature",
                                     "06": "Output voltage is too high",
                                     "07": "Over load time out",
                                     "08": "Bus voltage is too high",
                                     "09": "Bus soft start failed",
                                     "11": "Main relay failed",
                                     "51": "Over current inverter",
                                     "52": "Bus soft start failed",
                                     "53": "Inverter soft start failed",
                                     "54": "Self-test failed",
                                     "55": "Over DC voltage on output of inverter",
                                     "56": "Battery connection is open",
                                     "57": "Current sensor failed",
                                     "58": "Output voltage is too low",
                                     "60": "Inverter negative power",
                                     "71": "Parallel version different",
                                     "72": "Output circuit failed",
                                     "80": "CAN communication failed",
                                     "81": "Parallel host line lost",
                                     "82": "Parallel synchronized signal lost",
                                     "83": "Parallel battery voltage detect different",
                                     "84": "Parallel Line voltage or frequency detect different",
                                     "85": "Parallel Line input current unbalanced",
                                     "86": "Parallel output setting different"}],
            ["float", "L2 AC input voltage", "V"],
            ["float", "L2 AC input frequency", "Hz"],
            ["float", "L2 AC output voltage", "V"],
            ["float", "L2 AC output frequency", "Hz"],
            ["int", "L2 AC output apparent power", "VA"],
            ["int", "L2 AC output active power", "W"],
            ["int", "L2 Load percentage", "%"],
            ["float", "L2 Battery voltage", "V"],
            ["int", "L2 Battery charging current", "A"],
            ["int", "L2 Battery capacity", "%"],
            ["float", "PV2 Input Voltage", "V"],
            ["int", "PV2 charging current", "A"],
            ["flags", "Inverter Status", [
                "is_l2_scc_ok",
                "is_l2_ac_charging",
                "is_l2_scc_charging",
                "is_battery_over_voltage",
                "is_battery_under_voltage",
                "is_l2_line_lost",
                "is_l2_load_on",
                "is_configuration_changed"]]
        ],
        "test_responses": [
            b"(11 92911906100045 L 00 124.2 59.98 124.2 59.98 0149 0130 005 56.1 000 100 000.0 00 01000010\xA9\xA8\r",
        ],
        "regex": "QP2GS(\\d)$"
    },
    'QPGS': {
        "name": "QPGS",
        "description": "Parallel Information inquiry LV5048",
        "help": " -- example: QPGS1 queries the values of various metrics from instance 1 of parallel setup Inverters (numbers from 0)",
        "type": "QUERY",
        "response": [
                ["option", "Parallel instance number", ["Not valid", "valid"]],
                ["int", "Serial number", ""],
                ["keyed", "Work mode", {"P": "Power On Mode",
                                        "S": "Standby Mode",
                                        "L": "Line Mode",
                                        "B": "Battery Mode",
                                        "F": "Fault Mode",
                                        "H": "Power Saving Mode"}],
                ["keyed", "Fault code", {"00": "No fault",
                                         "01": "Fan is locked",
                                         "02": "Over temperature",
                                         "03": "Battery voltage is too high",
                                         "04": "Battery voltage is too low",
                                         "05": "Output short circuited or Over temperature",
                                         "06": "Output voltage is too high",
                                         "07": "Over load time out",
                                         "08": "Bus voltage is too high",
                                         "09": "Bus soft start failed",
                                         "11": "Main relay failed",
                                         "51": "Over current inverter",
                                         "52": "Bus soft start failed",
                                         "53": "Inverter soft start failed",
                                         "54": "Self-test failed",
                                         "55": "Over DC voltage on output of inverter",
                                         "56": "Battery connection is open",
                                         "57": "Current sensor failed",
                                         "58": "Output voltage is too low",
                                         "60": "Inverter negative power",
                                         "71": "Parallel version different",
                                         "72": "Output circuit failed",
                                         "80": "CAN communication failed",
                                         "81": "Parallel host line lost",
                                         "82": "Parallel synchronized signal lost",
                                         "83": "Parallel battery voltage detect different",
                                         "84": "Parallel Line voltage or frequency detect different",
                                         "85": "Parallel Line input current unbalanced",
                                         "86": "Parallel output setting different"}],
                ["float", "L1 AC input voltage", "V"],
                ["float", "L1 AC input frequency", "Hz"],
                ["float", "L1 AC output voltage", "V"],
                ["float", "L1 AC output frequency", "Hz"],
                ["int", "L1 AC output apparent power", "VA"],
                ["int", "L1 AC output active power", "W"],
                ["int", "L1 Load percentage", "%"],
                ["float", "Battery voltage", "V"],
                ["int", "Battery charging current", "A"],
                ["int", "Battery capacity", "%"],
                ["float", "PV1 Input Voltage", "V"],
                ["int", "Total charging current", "A"],
                ["int", "Total AC output apparent power", "VA"],
                ["int", "Total output active power", "W"],
                ["int", "Total AC output percentage", "%"],
                ["flags", "Inverter Status", [
                    "is_l1_scc_ok",
                    "is_l1_ac_charging_on",
                    "is_l1_scc_charging_on",
                    "is_battery_over_voltage",
                    "is_battery_under_voltage",
                    "is_l1_line_off",
                    "is_l1_load_on",
                    "is_configuration_changed"]],
                ["option", "Output mode", ["Standalone?",
                                           "Parallel output 0 degrees",
                                           "Phase 1 of 3 Phase output",
                                           "Phase 2 of 3 Phase output",
                                           "Phase 3 of 3 Phase output",
                                           "Parallel output 120 degrees",
                                           "Parallel output 180 degrees"]],
                ["option", "Charger source priority", ["Utility first", "Solar first", "Solar + Utility", "Solar only"]],
                ["int", "Max charger current", "A"],
                ["int", "Max charger range", "A"],
                ["int", "Max AC charger current", "A"],
                ["int", "PV1 charging current", "A"],
                ["int", "Battery discharge current", "A"]
        ],
        "test_responses": [
            b"(1 92911906100045 L 00 122.9 59.98 122.9 59.98 0331 0272 013 56.1 004 100 000.0 004 01577 01400 009 01000010 6 0 060 220 40 00 000\xC7\xC2\r",
        ],
        "regex": "QPGS(\\d)$",
    },
    'QPIGS': {
        "name": "QPIGS",
        "description": "General Status Parameters inquiry LV5048",
        "help": " -- queries the value of various metrics from the Inverter",
        "type": "QUERY",
        "response": [
                ["float", "L1 AC Input Voltage", "V"],
                ["float", "L1 AC Input Frequency", "Hz"],
                ["float", "L1 AC Output Voltage", "V"],
                ["float", "L1 AC Output Frequency", "Hz"],
                ["int", "L1 AC Output Apparent Power", "VA"],
                ["int", "L1 AC Output Active Power", "W"],
                ["int", "L1 AC Output Load", "%"],
                ["int", "BUS Voltage", "V"],
                ["float", "Battery Voltage", "V"],
                ["int", "Battery Charging Current", "A"],
                ["int", "Battery Capacity", "%"],
                ["int", "Inverter Heat Sink Temperature", "Deg_C"],
                ["float", "PV Input Current for Battery", "A"],
                ["float", "PV Input Voltage", "V"],
                ["float", "Battery Voltage from SCC", "V"],
                ["int", "Battery Discharge Current", "A"],
                ["flags", "Inverter Status", [
                    "is_l1_scc_ok",
                    "is_l1_ac_charging_on",
                    "is_l1_scc_charging_on",
                    "is_battery_over",
                    "is_battery_under",
                    "is_l1_line_not_ok",
                    "is_load_on",
                    "is_configuration_changed"]],
                ["int", "RSV1", "A"],
                ["int", "RSV2", "A"],
                ["int", "PV Input Power", "W"],
                ["flags", "Device Status2", [
                    "is_charging_to_float",
                    "is_switched_on",
                    "is_reserved"]]
        ],
        "test_responses": [
            b"(000.0 00.0 230.0 49.9 0161 0119 003 460 57.50 012 100 0069 0014 103.8 57.45 00000 00110110 00 00 00856 010\x24\x8C\r",
        ],
        "regex": "",
    },
    'QPIGS2': {
        "name": "QPIGS2",
        "description": "General Status Parameters inquiry",
        "help": " -- queries the value of various metrics from the Inverter",
        "type": "QUERY",
        "response": [
                ["float", "L2 AC Input Voltage", "V"],
                ["float", "L2 AC Input Frequency", "Hz"],
                ["float", "L2 AC Output Voltage", "V"],
                ["float", "L2 AC Output Frequency", "Hz"],
                ["int", "L2 AC Output Apparent Power", "VA"],
                ["int", "L2 AC Output Active Power", "W"],
                ["int", "L2 AC Output Load", "%"],
                ["int", "PV2 Battery Charging Current", "A"],
                ["float", "PV2 Input Voltage", "V"],
                ["float", "L2 Battery Voltage", "V"],
                ["flags", "Device Status", [
                    "is_l2_scc_ok",
                    "is_l2_ac_charging_on",
                    "is_l2_scc_charging_on",
                    "reserved",
                    "is_l2_line_not_ok",
                    "is_load_on",
                    "reserved"]]
        ],
        "test_responses": [
            b"",
        ],
        "regex": "",
    },
    'QPIRI': {
        "name": "QPIRI",
        "description": "Current Settings inquiry for LV5048",
        "help": " -- queries the current settings from the Inverter",
        "type": "QUERY",
        "response": [
                ["float", "AC Input Voltage", "V"],
                ["float", "AC Input Current", "A"],
                ["float", "AC Output Voltage", "V"],
                ["float", "AC Output Frequency", "Hz"],
                ["float", "AC Output Current", "A"],
                ["int", "AC Output Apparent Power", "VA"],
                ["int", "AC Output Active Power", "W"],
                ["float", "Battery Voltage", "V"],
                ["float", "Battery Recharge Voltage", "V"],
                ["float", "Battery Under Voltage", "V"],
                ["float", "Battery Bulk Charge Voltage", "V"],
                ["float", "Battery Float Charge Voltage", "V"],
                ["option", "Battery Type", ["AGM", "Flooded", "User"]],
                ["int", "Max AC Charging Current", "A"],
                ["int", "Max Charging Current", "A"],
                ["option", "Input Voltage Range", ["Appliance", "UPS"]],
                ["option", "Output Source Priority", ["Utility first", "Solar first", "SBU first"]],
                ["option", "Charger Source Priority", ["Utility first", "Solar first", "Solar + Utility", "Only solar charging permitted"]],
                ["int", "Max Parallel Units", "units"],
                ["keyed", "Machine Type", {"00": "Grid tie", "01": "Off Grid", "10": "Hybrid"}],
                ["option", "Topology", ["transformerless", "transformer"]],
                ["option", "Output Mode", ["Standalone?",
                                           "Parallel output 0 degrees",
                                           "Phase 1 of 3 Phase output",
                                           "Phase 2 of 3 Phase output",
                                           "Phase 3 of 3 Phase output",
                                           "Parallel output 120 degrees",
                                           "Parallel output 180 degrees"]],
                ["float", "Battery Redischarge Voltage", "V"],
                ["option", "PV OK Condition", ["As long as one unit of inverters has connect PV, parallel system will consider PV OK",
                                               "Only All of inverters have connect PV, parallel system will consider PV OK"]],
                ["option", "PV Power Balance", ["PV input max current will be the max charged current",
                                                "PV input max power will be the sum of the max charged power and loads power"]],
                ["int", "Max Charging Time at CV Stage", "min"]],
        "test_responses": [
            b"(230.0 21.7 230.0 50.0 21.7 5000 4000 48.0 46.0 42.0 56.4 54.0 0 10 010 1 0 0 6 01 0 0 54.0 0 1 60\x83\xAA\r",
            b"(120.0 20.8 120.0 60.0 20.8 2500 2500 48.0 47.0 43.0 56.1 56.1 2 40 060 0 0 0 9 01 0 6 55.0 0 1 000\xBE\xAC\r",
            b"(NAK\x73\x73\r",
        ],
        "regex": "",
    },
}


class pi41(pi30):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b'PI41'
        self.COMMANDS.update(NEW_COMMANDS)
        self.STATUS_COMMANDS = ['QPIGS', 'Q1']
        self.SETTINGS_COMMANDS = ['QPIRI', 'QFLAG']
        self.DEFAULT_COMMAND = 'QDI'
        # log.info(f'Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands')



import abc
import importlib
import logging

from ..helpers import get_kwargs
from ..io.testio import TestIO

log = logging.getLogger("MPP-Solar")

PORT_TYPE_UNKNOWN = 0
PORT_TYPE_TEST = 1
PORT_TYPE_USB = 2
PORT_TYPE_ESP32 = 4
PORT_TYPE_SERIAL = 8
PORT_TYPE_JKBLE = 16
PORT_TYPE_MQTT = 32


class AbstractDevice(metaclass=abc.ABCMeta):
    """
    Abstract device class
    """

    def __init__(self, *args, **kwargs):
        self._protocol = None
        self._protocol_class = None
        self._port = None
        log.debug(f"{self._classname} __init__ args {args}")
        log.debug(f"{self._classname} __init__ kwargs {kwargs}")
        self._name = get_kwargs(kwargs, "name")
        self.set_port(**kwargs)
        self.set_protocol(**kwargs)
        log.debug(
            f"{self._classname} __init__ name {self._name}, port {self._port}, protocol {self._protocol}"
        )

    def __str__(self):
        """
        Build a printable representation of this class
        """
        return f"{self._classname} device - name: {self._name}, port: {self._port}, protocol: {self._protocol}"

    def get_port_type(self, port):
        if port is None:
            return PORT_TYPE_UNKNOWN

        port = port.lower()
        # check for test type port
        if "test" in port:
            log.debug("device:get_port_type: port matches test")
            return PORT_TYPE_TEST
        # mqtt
        elif "mqtt" in port:
            log.debug("device:get_port_type: port matches mqtt")
            return PORT_TYPE_MQTT
        # USB type ports
        elif "hidraw" in port:
            log.debug("device:get_port_type: port matches hidraw")
            return PORT_TYPE_USB
        elif "mppsolar" in port:
            log.debug("device:get_port_type: port matches mppsolar")
            return PORT_TYPE_USB
        # ESP type ports
        elif "esp" in port:
            log.debug("device:get_port_type: port matches esp")
            return PORT_TYPE_ESP32
        # JKBLE type ports
        elif ":" in port:
            # all mac addresses currently return as JKBLE
            log.debug("device:get_port_type: port matches jkble ':'")
            return PORT_TYPE_JKBLE
        elif "jkble" in port:
            log.debug("device:get_port_type: port matches jkble")
            return PORT_TYPE_JKBLE
        elif "serial" in port:
            log.debug("device:get_port_type: port matches serial")
            return PORT_TYPE_SERIAL
        elif "ttyusb" in port:
            log.debug("device:get_port_type: port matches ttyusb")
            return PORT_TYPE_SERIAL
        else:
            return PORT_TYPE_UNKNOWN

    def set_protocol(self, *args, **kwargs):
        """
        Set the protocol for this device
        """
        protocol = get_kwargs(kwargs, "protocol")
        log.debug(f"device.set_protocol with protocol {protocol}")
        if protocol is None:
            self._protocol = None
            self._protocol_class = None
            return
        protocol_id = protocol.lower()
        # Try to import the protocol module with the supplied name (may not exist)
        try:
            proto_module = importlib.import_module("mppsolar.protocols." + protocol_id, ".")
        except ModuleNotFoundError:
            log.error(f"No module found for protocol {protocol_id}")
            self._protocol = None
            self._protocol_class = None
            return
        # Find the protocol class - classname must be the same as the protocol_id
        try:
            self._protocol_class = getattr(proto_module, protocol_id)
        except AttributeError:
            log.error(f"Module {proto_module} has no attribute {protocol_id}")
            self._protocol = None
            self._protocol_class = None
            return
        # Instantiate the class
        # TODO: fix protocol instantiate
        self._protocol = self._protocol_class("init_var", proto_keyword="value", second_keyword=123)

    def set_port(self, *args, **kwargs):
        port = get_kwargs(kwargs, "port")
        baud = get_kwargs(kwargs, "baud", 2400)
        porttype = get_kwargs(kwargs, "porttype")

        if porttype:
            log.info(f"Port overide: using port: {porttype}")
            port_type = self.get_port_type(porttype)
        else:
            port_type = self.get_port_type(port)

        if port_type == PORT_TYPE_TEST:
            log.info("Using testio for communications")
            from mppsolar.io.testio import TestIO

            self._port = TestIO()

        elif port_type == PORT_TYPE_USB:
            log.info("Using hidrawio for communications")
            from mppsolar.io.hidrawio import HIDRawIO

            self._port = HIDRawIO(device_path=port)

        elif port_type == PORT_TYPE_ESP32:
            log.info("Using esp32io for communications")
            from mppsolar.io.esp32io import ESP32IO

            self._port = ESP32IO(device_path=port)

        elif port_type == PORT_TYPE_JKBLE:
            log.info("Using jkbleio for communications")
            from mppsolar.io.jkbleio import JkBleIO

            self._port = JkBleIO(device_path=port)

        elif port_type == PORT_TYPE_SERIAL:
            log.info("Using serialio for communications")
            from mppsolar.io.serialio import SerialIO

            self._port = SerialIO(device_path=port, serial_baud=baud)

        elif port_type == PORT_TYPE_MQTT:

            mqtt_broker = get_kwargs(kwargs, "mqtt_broker", "localhost")
            mqtt_port = get_kwargs(kwargs, "mqtt_port", 1883)
            mqtt_user = get_kwargs(kwargs, "mqtt_user")
            mqtt_pass = get_kwargs(kwargs, "mqtt_pass")
            log.info(
                f"Using mqttio for communications broker {mqtt_broker}, port {mqtt_port}, user {mqtt_user}, pass {mqtt_pass}"
            )

            from mppsolar.io.mqttio import MqttIO

            self._port = MqttIO(
                client_id=self._name,
                mqtt_broker=mqtt_broker,
                mqtt_port=mqtt_port,
                mqtt_user=mqtt_user,
                mqtt_pass=mqtt_pass,
            )

        else:
            self._port = None

    def list_commands(self):
        # print(f"{'Parameter':<30}\t{'Value':<15} Unit")
        if self._protocol is None:
            log.error("Attempted to list commands with no protocol defined")
            return {"ERROR": ["Attempted to list commands with no protocol defined", ""]}
        result = {}
        result["_command"] = "command help"
        result[
            "_command_description"
        ] = f"List available commands for protocol {str(self._protocol._protocol_id, 'utf-8')}"
        for command in self._protocol.COMMANDS:
            result[command] = [self._protocol.COMMANDS[command]["description"], ""]
        return result

    def list_outputs(self):
        import pkgutil

        pkgpath = __file__
        pkgpath = pkgpath[: pkgpath.rfind("/")]
        pkgpath += "/../outputs"
        # print(pkgpath)
        result = {}
        result["_command"] = "outputs help"
        result["_command_description"] = "List available output modules"
        for _, name, _ in pkgutil.iter_modules([pkgpath]):
            # print(name)
            _module_class = importlib.import_module("mppsolar.outputs." + name, ".")
            _module = getattr(_module_class, name)
            # print(_module())
            result[name] = (str(_module()), "", "")
        # print(result)
        return result

    def run_commands(self, commands) -> dict:
        """
        Run multiple commands sequentially
        :param commands: List of commands to run, either with or without an alias.
        If an alias is provided, it will be substituted in place of cmd name in the returned dict
        Additional elements in the tuple will be passed to run_command as ordered
        Example: ['QPIWS', ...] or [('QPIWS', 'ALIAS'), ...] or [('QPIWS', 'ALIAS', True), ...] or mix and match
        :return: Dictionary of responses
        """
        responses = {}
        for i, command in enumerate(commands):
            if isinstance(command, str):
                responses[command] = self.run_command(command)
            elif isinstance(command, tuple) and len(command) > 0:
                if len(command) == 1:  # Treat as string
                    responses[command[0]] = self.run_command(command[0])
                elif len(command) == 2:
                    responses[command[1]] = self.run_command(command[0])
                else:
                    responses[command[1]] = self.run_command(command[0], *command[2:])
            else:
                responses["Command {:d}".format(i)] = {
                    "ERROR": ["Unknown command format", "(Indexed from 0)"]
                }
        return responses

    def run_command(self, command) -> dict:
        """
        generic method for running a 'raw' command
        """
        log.info(f"Running command {command}")
        if command == "list_commands":
            return self.list_commands()
        if command == "list_outputs":
            return self.list_outputs()
        if command == "get_status":
            return self.get_status()
        if command == "get_settings":
            return self.get_settings()
        if not command:
            command = self._protocol.DEFAULT_COMMAND

        if self._protocol is None:
            log.error("Attempted to run command with no protocol defined")
            return {"ERROR": ["Attempted to run command with no protocol defined", ""]}
        if self._port is None:
            log.error(f"No communications port defined - unable to run command {command}")
            return {
                "ERROR": [
                    f"No communications port defined - unable to run command {command}",
                    "",
                ]
            }

        # Send command and receive data
        full_command = self._protocol.get_full_command(command)
        log.info(f"full command {full_command} for command {command}")
        if full_command is None:
            log.error(
                f"full_command not found for {command} in protocol {self._protocol._protocol_id}"
            )
            return {
                "ERROR": [
                    f"full_command not found for {command} in protocol {self._protocol._protocol_id}",
                    "",
                ]
            }

        # Band-aid solution, need to reduce what is sent
        raw_response = self._port.send_and_receive(
            command=command,
            full_command=full_command,
            protocol=self._protocol,
            command_defn=self._protocol.get_command_defn(command),
        )
        log.debug(f"Send and Receive Response {raw_response}")

        # Handle errors; dict is returned on exception
        # Maybe there should a decode for ERRORs and WARNINGS...
        if isinstance(raw_response, dict):
            return raw_response

        # Decode response
        decoded_response = self._protocol.decode(raw_response, command)
        log.info(f"Decoded response {decoded_response}")

        return decoded_response

    def get_status(self) -> dict:
        # Run all the commands that are defined as status from the protocol definition
        data = {}
        for command in self._protocol.STATUS_COMMANDS:
            data.update(self.run_command(command))
        return data

    def get_settings(self) -> dict:
        # Run all the commands that are defined as settings from the protocol definition
        data = {}
        for command in self._protocol.SETTINGS_COMMANDS:
            data.update(self.run_command(command))
        return data



"""
A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open  # pylint: disable=redefined-builtin
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "pypi_README.md"), encoding="utf-8") as f:
    long_description = f.read()

exec(open(path.join(here, "mppsolar/version.py")).read())

setup(
    name="mpp-solar",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=__version__,  # noqa: F821
    description="Package to communicate with MPP Solar PIP-4048MS inverter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url="https://github.com/jblance/mpp-solar",
    # Author details
    author="John Blance",
    author_email="",
    # Choose your license
    license="MIT",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.6",
    # What does your project relate to?
    keywords="hardware serial communication",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["contrib", "daemon", "docs", "tests"]),
    include_package_data=True,
    data_files=[
        ("/etc/systemd/user", ["daemon/mpp-solar.service"]),
        ("/etc/systemd/user", ["daemon/jkbms.service"]),
        ("/etc/mpp-solar", ["daemon/mpp-solar.conf.example"]),
        ("/etc/jkbms", ["daemon/jkbms.conf.example"]),
    ],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    # install_requires=["pyserial", "paho-mqtt", "bluepy"],
    install_requires=["pyserial", "paho-mqtt"],
    # install_requires=[],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "dev": ["check-manifest"],
        "test": ["coverage"],
    },
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        "console_scripts": [
            "mpp-solar=mppsolar:main",
            "jkbms=mppsolar:main",
        ],
    },
    test_suite="tests.get_tests",
)



import logging

from .device import AbstractDevice
from ..io.jkbleio import JkBleIO

log = logging.getLogger("MPP-Solar")


class jkbms(AbstractDevice):
    def __init__(self, *args, **kwargs) -> None:
        self._classname = "jkbms"
        super().__init__(*args, **kwargs)

    def run_command(self, command) -> dict:
        """
        jkbms method for running a 'raw' command
        """
        log.info(f"JKBMS Running command {command}")
        # this is duplicated from parent
        if self._protocol is None:
            log.error("Attempted to run command with no protocol defined")
            return {"ERROR": ["Attempted to run command with no protocol defined", ""]}
        if self._port is None:
            log.error(f"No communications port defined - unable to run command {command}")
            return {
                "ERROR": [
                    f"No communications port defined - unable to run command {command}",
                    "",
                ]
            }

        # Send command and receive data
        full_command = self._protocol.get_full_command(command)
        log.info(f"full command {full_command} for command {command}")

        # JkBleIO is very different from the others, only has protocol jk02 and jk04, maybe change full_command?
        if isinstance(self._port, JkBleIO):
            # need record type, SOR
            raw_response = self._port.send_and_receive(command=command, protocol=self._protocol)

            log.debug(f"Send and Receive Response {raw_response}")

            # Handle errors; dict is returned on exception
            # Maybe there should a decode for ERRORs and WARNINGS...
            if isinstance(raw_response, dict):
                return raw_response

            # Decode response
            decoded_response = self._protocol.decode(raw_response, command)
            log.info(f"Decoded response {decoded_response}")
            return decoded_response

        else:
            return super().run_command(command)



import logging
from machine import UART
import time

from .baseio import BaseIO
from ..helpers import get_kwargs

log = logging.getLogger("MPP-Solar")


class ESP32IO(BaseIO):
    """
    Uses ESP32 ESP32 for  serial connection, sends command (multiple times if needed)
    and returns the byte_response
    """

    def __init__(self, device_path, serial_baud=2400) -> None:
        self._serial_port = device_path
        self._serial_baud = serial_baud

    def send_and_receive(self, *args, **kwargs) -> dict:
        full_command = get_kwargs(kwargs, "full_command")
        log.info(f"ESP32 serial connection: executing {full_command}")

        response_line = None
        uart_no = self._serial_port.lower().split("esp")[1]
        log.debug(f"port {self._serial_port}, baudrate {self._serial_baud}, uart# {uart_no}")
        try:
            with UART(uart_no, self._serial_baud) as s:
                # Execute command multiple times, increase timeouts each time
                s.init(self._baud_rate, timeout=1000)
                s.write(full_command)
                time.sleep(0.5)  # give serial port time to receive the data
                response_line = s.readline()
                log.debug("esp32 serial response was: %s", response_line)
                return response_line
        except Exception as e:
            log.warning("ESP32 Serial read error: {}".format(e))
        log.info("Command execution failed")
        return {"ERROR": ["Serial command execution failed", ""]}
